/* ez-blob.h: Serialising description and serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>
#include <stdio.h>

#include "ez-blob.h"
#include "ez-list.h"

#include "ez-blob-io.h"

void ez_blob_print(const ez_blob_desc *d, const void *a, int depth) {
	char x[depth+1];
	int len = d->bd_length;

	memset(x, ' ', depth);
	x[depth] = 0;

	for (int i=0;i<=len;i++,d++) {
		const void *u = a + d->bd_offset;

		printf("%s [type=$%04x, offset=$%04x]", x, d->bd_type, d->bd_offset);
		
		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			printf(" .%d = %02x\n", d->bd_id, *(uint8_t *)u);
			break;
		case EZ_BLOB_INT16:
			printf(" .%d = %04x\n", d->bd_id, *(uint16_t *)u);
			break;
		case EZ_BLOB_INT32:
			printf(" .%d = %08x\n", d->bd_id, *(uint32_t *)u);
			break;
		case EZ_BLOB_INT64:
			printf(" .%d = %016lx\n", d->bd_id, *(uint64_t *)u);
			break;
		case EZ_BLOB_FLOAT32:
			printf(" .%d = %f\n", d->bd_id, *(float *)u);
			break;
		case EZ_BLOB_FLOAT64:
			printf(" .%d = %f\n", d->bd_id, *(double *)u);
			break;
			
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP | EZ_BLOB_ISNULLABLE:
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
			u = *(void **)u;
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
			if (u) {
				printf(" .%d %p = {\n", d->bd_id, u);
				ez_blob_print(d->bd_table, u, depth+4);
				printf("%s  }\n", x);
			} else {
				printf(" .%d %p\n", d->bd_id, u);
			}
			break;

		case EZ_BLOB_CSTRING | EZ_BLOB_INT8 | EZ_BLOB_ISNULLABLE:
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8:
			printf(" .%d %p = `%s'\n", d->bd_id, *((char **)u), *((char **)u));
			break;

		case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64: {
			const struct ez_blob *ua = u;
			unsigned int elsize = 1 << (d->bd_type & EZ_BLOB_SIZE);
			ez_blob blob = { .eb_size = ua->eb_size * elsize, .eb_data = ua->eb_data };
			
			printf(" .%d %p [%zd x %d] = {\n", d->bd_id, ua->eb_data, ua->eb_size, elsize);
			ez_blob_dump(&blob, x);
			printf("%s  }\n", x);
			break;
		}

		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP | EZ_BLOB_ISNULLABLE:
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
			u = *(const void **)u;
		case EZ_BLOB_STRUCT | EZ_BLOB_LIST:
			if (u) {
				printf(" .%d %p [] = {\n", d->bd_id, u);
				for (ez_node *w = ez_list_head((ez_list *)u), *n = ez_node_next(w); n; w = n, n=ez_node_next(n)) {
					printf("%s{\n", x);
					ez_blob_print(d->bd_table, w, depth+4);
					printf("%s},\n", x);
				}
				printf("%s  }\n", x);
			} else {
				printf(" .%d %p\n", d->bd_id, u);
			}
			break;

		case EZ_BLOB_MAGIC:
			printf(" OBJECT id=%d length=%d\n", d->bd_id, d->bd_length);
			break;
		default:
			printf(" .%d ?? UNKNOWN ??\n", d->bd_id);
			break;
		}
	}
}

