/* ez-port.c: A simple inter-thread IPC mechanism with poll() support.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <fcntl.h> 
#include <sys/stat.h>  
#include <mqueue.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ez-port.h"

/**
 * This implementation uses anonymous posix message queues.
 *
 * Internally it uses the message queue to store the message pointers
 * directly.  This should work in-thread.  It might do funny things
 * with valgrind and the like since they pointers escape the user
 * address space.
 *
 * If two readers are reading at the time time, poll() may currently
 * block as it doesn't use non-blocking i/o.
 */

#define EZ_PORT_NAME "ez-port"

struct ez_port {
	mqd_t q;
};

ez_port *ez_port_new(int limit) {
	ez_port *port = malloc(sizeof(*port)+ sizeof(void *) * limit);
	struct mq_attr at;
	char name[16 + strlen(EZ_PORT_NAME) + 4];

	if (!port)
		return port;

	// Port name includes uid to avoid permission issues if a stale port exists
	sprintf(name, "/%s-%016lx", EZ_PORT_NAME, getuid() & 0xffffffffffffffffUL);

	at.mq_flags = 0;
	at.mq_maxmsg = limit;
	at.mq_msgsize = sizeof(void *);

	// Open a unqiue port anonymously, using O_EXCL and unlink().
	printf("open %s\n", name);
	do {
		mq_unlink(name);
		port->q = mq_open(name, O_RDWR | O_CREAT | O_EXCL | O_CLOEXEC, 0600, &at);
	} while (port->q == -1 && errno == EEXIST);

	if (port->q == -1) {
		perror("mq_open");
		free(port);
		return NULL;
	}
	mq_unlink(name);

	return port;
}

void ez_port_free(ez_port *port) {
	if (port) {
		mq_close(port->q);
		free(port);
	}
}

int ez_port_fd(ez_port *port) {
	return port->q;
}

int ez_port_put(struct ez_port *port, void *msg) {
	return mq_send(port->q, (char *)&msg, sizeof(msg), 0);
}

void *ez_port_poll(ez_port *port) {
	struct mq_attr at;
	
	if (mq_getattr(port->q, &at) == 0 && at.mq_curmsgs > 0)
		return ez_port_take(port);

	return NULL;
}

void *ez_port_take(ez_port *port) {
	void *tail;
	ssize_t size;

	size = mq_receive(port->q, (char *)&tail, sizeof(tail), NULL);
	if (size == sizeof(tail))
		return tail;

	return NULL;
}
